#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#define MY_INFINITY ((1UL << (sizeof(unsigned long)*8 -1 ))-1);



#ifndef __MEDIAN_STRING_
#define __MEDIAN_STRING_

//Size of the l-motif
extern unsigned char l_motif_size;
//#Of sequences
extern long seq_num;

//Pointer to the set of sequences
typedef struct SEQ{
	char *_sequence;
	char _seq_info[80];
	unsigned long len;
	void *_lease; //We use this to link any thing to this structure
    char *_sequence_end;
	struct SEQ *next;	
} Sequence;

extern Sequence *sequences;
extern unsigned long currentBest;
typedef enum __operation__ { EXPAND, PRUNE} Operation;

#endif

/*Golbal definitions*/
Sequence *sequences = NULL;
unsigned long currentBest = MY_INFINITY; 
long seq_num=-1;
long lmer_len=0;
static long current_index=0;
/*Neucleotide*/
typedef enum _ntide_ {A,T,G,C} Neucleotide;
Neucleotide *currentBest_N=NULL;
Neucleotide *currentBest_N_mismatch=NULL;
unsigned char is_prunable=0;

// 
// The function to parse the FASTA format and build the 
// sequences structure, Builds from any stream.
//

int buildFastaSequences(FILE *stream){

 char fasta_buf[80]; //No line is more than 80 char
 Sequence **ptr = &sequences;
 unsigned long seq_len = 0;
 char *sptr,*tmp_ptr;
 	
	seq_num = -1;

    /*We dont want anything in the stream otherwise [^\n] gets in trouble*/
	
	while(fscanf(stream,"%[^\n]",fasta_buf)!=EOF){
	    switch(fasta_buf[0]){
			case '>':
				/*New Sequence*/
				if(!*ptr){
					*ptr = (Sequence *)malloc(sizeof(Sequence)*1);
					if(!sscanf(fasta_buf,"%s",(*ptr)->_seq_info)){
						fprintf(stderr ,
								"Warning: Unable to scan sequence info for Seq# %d",seq_num);
					}

				} else {
					assert((*ptr)->next==NULL);
					(*ptr)->next = (Sequence *) malloc(sizeof(Sequence)*1);

                    if(!((*ptr)->next)){
                        fprintf(stderr,"Error: Unable to expand sequence %d\n",seq_num);
                        return -1;
                    }

					if(!sscanf(fasta_buf,"%s",((*ptr)->next)->_seq_info)){
						fprintf(stderr,
								"Warning: Unable to scan sequence info for Seq# %d",seq_num);
							
					}
					ptr = &((*ptr)->next);
				}

#ifdef __DEBUG__
					printf("STARTING NEW SEQUENCE %s\n",(*ptr)->_seq_info);
#endif
	     		++seq_num;
				(*ptr)->_sequence = NULL;
				(*ptr)->next = NULL;
				break;

			case ';':
				/*Ignore comment
				 */
				break;

			default:
				if(!(*ptr)){
					fprintf(stderr,
							"Error: Invalid format..., Expecting > or ; 	before anything\n");
					return -1;
				}
				/*Scan the sequence*/
				if(!(*ptr)->_sequence) {
					(*ptr)->_sequence = (char *) 
						calloc(strlen(fasta_buf)+1,sizeof(char));
                    *((*ptr)->_sequence)='\0';
				} else {
                    tmp_ptr = (char *) calloc((strlen(fasta_buf)
                                +strlen((*ptr)->_sequence))+1,sizeof(char));
					if(!tmp_ptr){
						fprintf(stderr,"Error: malloc failed...mallocing %ld\n",
								strlen(fasta_buf));
						perror("REASON::");
						return -1;
					}
                    sptr = (*ptr)->_sequence;
                    if(memcpy(tmp_ptr,sptr,strlen(sptr))!= tmp_ptr){
                        fprintf(stderr,"Error: memcpy failed while expanding the string\n");
                    }
                    /*ZAP a '\0' end of the sequence*/
                    *(tmp_ptr+strlen(sptr))='\0';

                    free((*ptr)->_sequence);
                    (*ptr)->_sequence = tmp_ptr;
				}
				/*MEMCPY*/
				sptr = (*ptr)->_sequence;
				assert(sptr != NULL);
				if(!sscanf(fasta_buf,"%s",(char *)((char *)sptr+strlen(sptr)))){
					fprintf(stderr,"Error: Error while scanning sequence %d\n",seq_num);
					return -1;
				}
                (*ptr)->_sequence_end = (char *)((*ptr)->_sequence + strlen(sptr));
                if(*((*ptr)->_sequence_end) != '\0'){
                    fprintf(stderr,"Error: Format Error\n");
                    return -1;
                }

#ifdef __DEBUG__
				printf("The current ptr sequence is %s\n",
						(*ptr)->_sequence);
#endif

		}
        getchar();
	}
	return 0;	
}

void printFastaSequences(){
	Sequence *ptr = sequences;
	while(ptr){
		printf(">%s\n",ptr->_seq_info);
		printf("%s\n",ptr->_sequence);
		ptr = ptr->next;
	}
}

#ifdef __UNIT_TEST_FASTA__
int main(){

	printf("Please Enter the sequences in FASTA format\n");
	
	if(buildFastaSequences(stdin)){
		printf("Problem in building the FASTA sequences\n");
				
	}

	printFastaSequences();

}
#endif

signed char BnBMedianString();
signed char char_compare(Neucleotide n,char c);
static int readLmerLength(FILE *stream){
    if(fscanf(stream,"%d",&lmer_len)!=1){
        fprintf(stderr,"Error: Failed to read L-mer size");
        return -1;
    }
}

Neucleotide  * prefixGenerator(Operation o);
void printNeucleotide(Neucleotide);
#ifdef __UNIT_TEST_PREFIX__

int main(){
    /**/
    Neucleotide *ptr=NULL;
    unsigned long i;
    current_index=-1;
    printf("Please enter l-mer length\n");
    scanf("%d",&lmer_len);
    while((ptr=prefixGenerator(EXPAND))){

        /*Testing prune*/
        if(ptr[0]==A && ptr[1]==T){
            ptr = prefixGenerator(PRUNE);
        }

        for(i=0;i<=current_index;i++){
            printNeucleotide(ptr[i]);
        }
        fprintf(stdout,"\n");
        
    }

}
#endif

void printNeucleotide(Neucleotide t){

    if(t == A){
        fprintf(stdout,"a");

    } else if(t == T){
        fprintf(stdout,"t");

    } else if(t == G){
        fprintf(stdout,"g");
    } else {
        fprintf(stdout,"c");
    }
}

Neucleotide  * prefixGenerator(Operation o){

    static Neucleotide *prefix =NULL;

    assert(lmer_len > 1);
    if(!prefix){
        prefix = (Neucleotide *) malloc(sizeof(Neucleotide)*lmer_len);
        if(!prefix){
            fprintf(stderr,"Error: Unable to create initial prefix\n");
            return NULL;
        }
    }

    if(o == EXPAND){
EXPAND_LABEL:
        if(current_index + 1 > lmer_len-1){
            /*You cannot expand now*/
            goto PRUNE_LABEL;
        } else {
            current_index ++ ;
            prefix[current_index] = A;
        }
    } else {

PRUNE_LABEL:
        do{
            if(prefix[current_index] == C){
              current_index--;
                if(current_index < 0) {
#ifdef __DEBUG__
                 fprintf(stderr,
                        "Warning: Searching the tree is complete exiting\n");
#endif
                 return NULL;
                }
                continue;
            } else {
                prefix[current_index]++;
                break;
            }
        }while(current_index>=0);

    }
    return prefix;

}


/*
 * Sets up the solution state: The state of the
 * solution is nothing but a index_array[seq_num].
 * Array of indices in the sequence array.
 *
 **/

unsigned long getTotalHammingDistance(Neucleotide *pre, unsigned long len);
int buildSolutionState(){
    
    if(fscanf(stdin,"%d", &lmer_len) < 1){
        fprintf(stderr,"Error: Unable to scan the l-mer length\n");
        return -1;
    }
    getchar();

    currentBest_N = (Neucleotide *) malloc(sizeof(Neucleotide)*lmer_len);
    if(!currentBest_N)
        return -1;
    currentBest_N_mismatch = (Neucleotide *) malloc(sizeof(Neucleotide)*lmer_len);
    if(!currentBest_N_mismatch)
        return -1;

    /*Build the FASTA sequences*/
    if(buildFastaSequences(stdin)<0){
        fprintf(stderr,"Error: Scanning FASTA format\n");
        return -1;
    }
}

unsigned long getMinHammingDistance(Sequence *,
        Neucleotide *prefix,unsigned long) ;

unsigned long getTotalHammingDistance(Neucleotide *pre, unsigned long len){
    Sequence *ptr = sequences;
    unsigned long hamming_total=0;
    char stop_prune = 1;
    is_prunable = 1;
    
    while(ptr){
        hamming_total += getMinHammingDistance(ptr,pre,len);
        ptr = ptr->next;
        stop_prune = stop_prune & is_prunable;
    }
    is_prunable = stop_prune;
    /*Add the score to the hamming_total*/
    assert(hamming_total >= 0);
    return hamming_total;
}

/*current_index is the length of the prefix.*/
int getMismatchScore(){

    Sequence *ptr = sequences;
    int score_table[4];
    int temp_score;
    unsigned long current_score = 0,total_score=0;
    unsigned long i,j,k;

    for(i=1;i<(lmer_len - current_index); i++){
        score_table[0]=0;score_table[1]=0;
        score_table[2]=0;score_table[3]=0;
        ptr = sequences;
        current_score=0;
        while(ptr){

         assert(ptr->_lease != NULL);

         if(char_compare(A,*(((char *)ptr->_lease)+i))){
            score_table[0]++;
         } else if(char_compare(T,*(((char *)ptr->_lease)+i))){
            score_table[1]++;
         } else if(char_compare(G,*(((char *)ptr->_lease)+i))){
             score_table[2]++;
         } else 
             score_table[3]++;
         ptr = ptr->next;
        }
        /*Maxmimum repeating entry in score_table from 0..4 is 
         the consensus*/
        temp_score =0;
        for(k=0;k<4;k++){
            if(score_table[k]>score_table[temp_score]){
                temp_score = k;
            }
            /*Mismatch Score, Ideally we want it 0*/
            current_score += score_table[k];
        }
        total_score += current_score - score_table[temp_score];
        currentBest_N_mismatch[current_index+i] = A + temp_score;
    }
    return total_score;

}

signed char char_compare(Neucleotide n,char c){

    switch(c){

        case 'A':
        case 'a':
            if(n!=A)return 0;
            break;

        case 'T':
        case 't':
            if(n!=T) return 0;
            break;

        case 'G':
        case 'g':
            if(n!=G) return 0;
            break;

        case 'c':
        case 'C':
            if(n!=C) return 0;
            break;


    }
    return 1;
    
}

unsigned long getMinHammingDistance(Sequence *ptr,Neucleotide *prefix, 
        unsigned long len){
    char *index_ptr = ptr->_sequence;
    char *index_ptr_end = ptr->_sequence_end;
    char *score_ptr = NULL;
    unsigned long i;
    unsigned long curr_len,opt_len;
    opt_len = MY_INFINITY;
    do{
        curr_len = 0;
        for(i=0;i<=len;i++){

          if(!char_compare(prefix[i],*(index_ptr+i))){
              curr_len++;
          }  
        }

       if(curr_len < opt_len){
            opt_len = curr_len;
            ptr->_lease = index_ptr+len;
            is_prunable = 1;
       } else if(curr_len == opt_len){
           is_prunable = 0;
       }

       index_ptr++;
    }while((index_ptr_end - index_ptr)>=lmer_len);
   
    return opt_len;
}

int main(){

     BnBMedianString();
}


signed char BnBMedianString(){

    Neucleotide *ptr=NULL;
    unsigned long i,j;
    unsigned long minDistance;
    unsigned long mismatchScore;
    Operation o = EXPAND;
    if(buildSolutionState()<0){
        fprintf(stderr,"BnBMedianStringAlgo: Failed during initialization\n");
        return -1;
    }
    current_index=-1;
    while((ptr=prefixGenerator(o))){
        minDistance = getTotalHammingDistance(ptr,current_index);

        if(current_index < lmer_len-1){

            if(is_prunable){
             mismatchScore = (unsigned long)getMismatchScore();
            }
            else
             mismatchScore = 0;

             if(minDistance + mismatchScore > currentBest){
                 if(mismatchScore && minDistance < currentBest ){
                     printf("Extra subtree pruned at::mismatchScore=%ld,currentBest=%ld,%ld\n",
                     mismatchScore,currentBest,minDistance+mismatchScore);
                    for(i=0;i<=current_index;i++){
                     printNeucleotide(ptr[i]);
                    }
                     printf("\n");
                 }
                o = PRUNE;
             }
             else{
                o = EXPAND;
             }

        } else {
            if(minDistance < currentBest){
                currentBest = minDistance;
                for(i=0;i<=current_index;i++){
                  currentBest_N[i] = ptr[i];
                }
            }
            o = EXPAND;
        }
    }

    printf("Median: ");
    for(i=0;i<lmer_len;i++)
        printNeucleotide(currentBest_N[i]);
    printf("\n");
    printf("TotalDistance: %ld\n",currentBest);

    return 0;
}
